; Copyright 2003 (c) Kevin Atkinson under the GNU GPL license version
; 2.0.	You should have received a copy of the GPL license along with
; this program if you did not you can find it at http://www.gnu.org/.

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Macros
; 

%macro MMCON 2-3 eax ; reg, con, treg: makes mmx constant, uses eax by default
	mov  %3, %2
	movd %1, %3
	punpckldq %1,%1
%endmacro

%macro XMMCON 2-3 eax ; reg, con makes xmmx constant, uses eax by default
	mov  %3, %2
	movd %1, %3
	pshufd %1,%1,0x00
%endmacro

%macro SWAPM 2 ; swap memory relative to ebp, uses eax, ebx
	mov  eax, [ebp + %1]
	mov  ebx, [ebp + %2]
	mov  [ebp + %2], eax
	mov  [ebp + %1], ebx
%endmacro

%macro TIME_B 1 ; if loc is esp must add 8
	sub esp,8
	mov [esp + 0], eax
	mov [esp + 4], edx
	rdtsc
	mov %1, eax
	mov eax, [esp + 0]
	mov edx, [esp + 4]
	add esp,8
%endmacro

%macro TIME_E 1 ; clears mmx reg, ebp must point to CompareParms
	sub esp,8
	mov [esp + 0], eax
	mov [esp + 4], edx
	emms
	rdtsc
	mov edx, %1
	sub  eax, edx
	mov  edx, ebp
	call _print_cs
	mov eax, [esp + 0]
	mov edx, [esp + 4]
	add esp,8
%endmacro

; WILL effect status registers
%macro align16 0
	times (($$-$)&15)/12 add dword esi, strict dword 0
	times (($$-$)&15)/6  or  dword edi, strict dword 0
	times (($$-$)&15)/3  add dword ebx, strict byte 0
	times (($$-$)&15)/2  or ecx, ecx
	times (($$-$)&15)    nop
%endmacro

	section .text

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Useful constants
;
struc common_data ; .* fields are not accessed
	.cpu		resd 1
	.cf_row		resd 1
	cf_row_loop     resd 1
	cf_mul		resd 1
	row_size	resd 1
	s_height	resd 1
	s_width		resd 1
	s_pitch		resd 1
	.data		resd 1
	scale		resd 1
	.log		resd 1
	a		resd 1
	b		resd 1
	c		resd 1
	d		resd 1
	e		resd 1
	prev0		resd 1
	cur0		resd 1
	next0		resd 1
	prev1		resd 1
	cur1		resd 1
	next1		resd 1

	cp_size

	s_l_0		resd 4
	s_l_1		resd 4

	line		resd 1
endstruc
struc	s_l
	s_l_data	resd 1
	s_l_stop	resd 1
	s_l_pitch	resd 1
	s_l_subsample_line resd 1
endstruc
	
	s_size		equ	common_data_size
      
	parms		equ	s_size + 20
	s_l_orig_0	equ	s_size + 24
	s_l_orig_1	equ	s_size + 28

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; _asm_subsample_line
;
; Avoids double function call by jumping
;
; ebp = data ptr, eax = dest, esi = i
; clobbers eax, ebx, ecx, edx
;
	align 16
	global _asm_subsample_line 
_asm_subsample_line:
	mov ebx, esi
	shl ebx, 4
	lea ebx, [ebp + ebx + s_l_0]

	mov ecx, [ebx + s_l_data]
	add ecx, [ebx + s_l_pitch]

	mov [ebx + s_l_data], ecx
	
	cmp ecx, [ebx + s_l_stop]

	mov edx, [ebp + s_width]
	lea edx, [eax + 2*edx]
	jnl _memclear
	
	mov edx, eax
	mov eax, ecx
	mov ecx, [ebp + row_size]
	
	jmp [ebx + s_l_subsample_line]

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; _subsample_line_*_2_gen
;
;

%macro GEN_SUBSAMPLE_LINE_GEN 2

	global _subsample_line_ %+ %1 %+ _2_gen
	align 16
; regparm(3) calling convention
_subsample_line_ %+ %1 %+ _2_gen:
	push	ebx
	and	ecx, ~(2*%2 - 1) ; floor
	push	esi
	xor	ebx, ebx
	push	edi
	lea	esi, [eax + 4*%2]
	lea	edi, [ecx + eax]
	align16
.loop:	
	; esi is x + 4*%2 to avoid going out of bounds when row_size
	; is not a multiple of 4*%2
	xor	ecx, ecx
	xor	eax, eax
	mov	cl, byte [esi     -4*%2]	    
	mov	al, byte [esi+1*%2-4*%2]
	add	cx, bx
	add	cx, ax
	mov	word [edx], cx
	xor     ecx, ecx
	xor     ebx, ebx
	mov	cl, byte [esi+2*%2-4*%2]
	mov	bl, byte [esi+3*%2-4*%2]
	add     cx, ax
	add     cx, bx
	add	esi, 4*%2
	mov	word [edx+2], cx
	add	edx, 4
	cmp	esi, edi
	jle	.loop
	sub	esi, 4*%2
	cmp	esi, edi
	pop	edi
	jne	.cleanup
	pop	esi
	pop	ebx
	ret
	align 16
.cleanup	
	xor	ecx, ecx
	xor	eax, eax
	mov	cl, byte [esi     ]	    
	mov	al, byte [esi+1*%2]
	add	cx, bx
	add	cx, ax
	mov	word [edx], cx
	pop	esi
	pop	ebx
	ret
%endmacro

GEN_SUBSAMPLE_LINE_GEN yuy2, 2
GEN_SUBSAMPLE_LINE_GEN yv12, 1

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

%macro NOW_YUY2 0
	%define WHAT yuy2
	%define MMWIDTH 2*MWIDTH
%endmacro
%macro NOW_YV12 0
	%define WHAT yv12
	%define MMWIDTH MWIDTH
%endmacro	 
	
%macro SS_MM_PRE 0
	prefetchnta [eax]
	push  ebx
	lea   ebx, [eax + MWIDTH - 1]
	and   ebx, ~(MWIDTH - 1)     ; ebx = ceil(eax)
	sub   ebx, eax               ; ebx = ceil(eax) - eax
	sub   ecx, ebx               ; ecx = ecx - ebx
	and   ecx, ~(MMWIDTH - 1)    ; ecx = floor(ecx)
	add   eax, ebx               ; eax = eax + ebx 
				     ;     = eax + ceil(eax) - eax
				     ;     = ceil(eax)
	add   ecx, eax               ; ecx (now)= stop
%endmacro

%macro SS_MM_POST 0
	pop ebx
%endmacro

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; _subsample_line_*_2_isse
;
; uses regparm(3) calling convention
;

%define MWIDTH 8

%macro SS_ISSE_COMMON 0
	movq  mm3, mm2
	pand  mm2, mm4
	psrlq mm3, 8
	pand  mm3, mm4
	movq  mm1, mm3
	psllq mm1, 16
	paddw mm2, mm3
	por   mm1, mm0
	paddw mm2, mm1

	movq [edx], mm2
	
	movq  mm0, mm3
	psrlq mm0, 48 
%endmacro

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
		
	global _subsample_line_yuy2_2_isse
	NOW_YUY2
	align 16
_subsample_line_yuy2_2_isse:
	SS_MM_PRE

	pxor  mm0, mm0
	MMCON mm4,0x00FF00FF, ebx

	align16
.loop:
	prefetchnta [eax + 16]

	movq  mm2, [eax    ]
	movq  mm1, [eax + 8]

	pand  mm2, mm4
	pand  mm1, mm4
	
	packuswb mm2,mm1

	SS_ISSE_COMMON
	
	add   eax, 16
	add   edx, 8

	cmp   eax, ecx
	jl    .loop

	SS_MM_POST

	ret
	
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

	global _subsample_line_yv12_2_isse
	NOW_YV12
	align 16
_subsample_line_yv12_2_isse:
	SS_MM_PRE

	pxor mm0, mm0
	MMCON mm4,0x00FF00FF, ebx

	align16
.loop:
	prefetchnta [eax + 8]

	movq  mm2, [eax]

	SS_ISSE_COMMON

	add	eax, 8
	add	edx, 8

	cmp	eax, ecx
	jl	.loop

	SS_MM_POST

	ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; _subsample_line_*_2_isse2
;
; uses regparm(3) calling convention
;

%define MWIDTH 16

%macro SS_ISSE2_COMMON 0
	movdqa xmm3, xmm2
	pand   xmm2, xmm4
	psrldq xmm3, 1
	pand   xmm3, xmm4
	movdqa xmm1, xmm3
	pslldq xmm1, 2
	paddw  xmm2, xmm3
	por    xmm1, xmm0
	paddw  xmm2, xmm1

	movdqa [edx], xmm2
	
	movdqa xmm0, xmm3
	psrldq xmm0, 14 
%endmacro

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

	global _subsample_line_yuy2_2_isse2
	NOW_YUY2
	align 16
_subsample_line_yuy2_2_isse2:
	SS_MM_PRE

	pxor  xmm0, xmm0
	XMMCON xmm4,0x00FF00FF, ebx

	align16
.loop:
	prefetchnta [eax + 32]

	movdqa xmm2, [eax     ]
	movdqa xmm1, [eax + 16]

	pand  xmm2, xmm4
	pand  xmm1, xmm4
	
	packuswb xmm2,xmm1

	SS_ISSE2_COMMON
	
	add   eax, 32
	add   edx, 16

	cmp   eax, ecx
	jl    .loop

	SS_MM_POST

	ret
	
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

	global _subsample_line_yv12_2_isse2
	NOW_YV12
	align 16
_subsample_line_yv12_2_isse2:
	SS_MM_PRE

	pxor xmm0, xmm0
	XMMCON  xmm4,0x00FF00FF, ebx

	align16
.loop:
	prefetchnta [eax + 16]

	movdqa  xmm2, [eax]

	SS_ISSE2_COMMON

	add	eax, 16
	add	edx, 16

	cmp	eax, ecx
	jl	.loop

	SS_MM_POST

	ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; _compare_fields_sse
;
; normal C calling convention
;
	global _compare_fields_sse
	align 16
_compare_fields_sse:			    
	push  ebp
	push  edi
	push  esi
	push  ebx

	sub   esp, s_size
	mov   ebp, esp ; ebp = working data ptr, entire file

	mov   eax, [ebp + parms]
	lea   ebx, [eax + cp_size]
	mov   edx, ebp
	call  _asm_memcopy
	
	mov   eax, [ebp + s_l_orig_0]
	lea   ebx, [eax + s_l_size]
	lea   edx, [ebp + s_l_0]
	call  _asm_memcopy

	mov   eax, [ebp + s_l_orig_1]
	lea   ebx, [eax + s_l_size]
	lea   edx, [ebp + s_l_1]
	call  _asm_memcopy
	
	mov   [ebp + line], dword 0  
	
	mov   esi, [ebp + s_pitch]
	
	mov   eax, [ebp + prev0]
	lea   edx, [eax + 2*esi]
	call  _memclear
	
	mov   eax, [ebp + prev1]
	lea   edx, [eax + 4*esi]
	call  _memclear
	
	mov   eax, [ebp + d]
	mov   esi, 0
	call  _asm_subsample_line
	
	mov   eax, [ebp + e]
	mov   esi, 1
	call  _asm_subsample_line

	xorps xmm7, xmm7 ; xmm7 = accum [entire function]
	
.outer:
	mov   edx, [ebp + cur1]
	call  [ebp + cf_mul]
	
	mov   edx, [ebp + next1]
	call  [ebp + cf_mul]
	
	mov   eax, [ebp + prev1]
	mov   ebx, [ebp + cur1]
	mov   ecx, [ebp + next1]
	
	mov   esi, 0
	mov   edi, [ebp + s_width]
	shl   edi, 2
	
	align 16
.inner:
	movaps xmm0, [eax + esi]
	addps  xmm0, [ecx + esi]
	mulps  xmm0, [ebx + esi]
	addps  xmm7, xmm0
	add    esi, 16
	cmp    esi, edi
	jl     .inner
	
	SWAPM prev1, next1

	; outer loop test
	mov	eax, [ebp + s_l_0 + s_l_data]
	cmp	eax, [ebp + s_l_0 + s_l_stop]
	jl	.outer

	mov	ebx, [ebp + s_l_1 + s_l_data]
	cmp	ebx, [ebp + s_l_1 + s_l_stop]
	jl	.outer

	emms

	movhlps xmm0, xmm7 ; sum up 4 floats in xmm7
	addps	xmm0, xmm7
	movaps	xmm1, xmm0
	shufps	xmm1, xmm1, 0x01
	addss	xmm0, xmm1
	mulss	xmm0, [ebp + scale]
	movss	[ebp+0], xmm0

	fld	dword [ebp+0]

	add	esp, dword s_size
	pop	ebx
	pop	esi
	pop	edi
	pop	ebp
	ret
	

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; compare_fields_mul_sse
;
; edx = dest, esp = common data ptr
; will clobber all gp registers except ebp, esp, all mmx registers,
; and all sse registers except xmm7
;
	global _asm_compare_fields_mul_sse
	align 16
_asm_compare_fields_mul_sse: 
	push edx

	mov   eax, ebp	  
	mov   edx, [ebp + cur0]
	call  _compare_fields_row_isse
	
	mov   eax, ebp
	mov   edx, [ebp + next0]
	call  _compare_fields_row_isse
	
	mov edx, [esp]
	       
	MMCON mm0, 0x00000000
		
	mov  eax, [ebp + prev0]
	mov  ebx, [ebp + cur0]
	mov  ecx, [ebp + next0]
	mov  esi, 0
	mov  edi, [ebp + s_width]
	shl  edi, 1
	
	align16
.loop:
	movq	  mm1, [eax + esi]
	movq	  mm2, [ecx + esi]
	paddusw	  mm1, mm2
	
	movq	  mm2, mm1
	punpckhwd mm1, mm0  ; high 
	cvtpi2ps xmm1, mm1 
	movlhps	 xmm1, xmm1
	punpcklwd mm2, mm0  ; low
	cvtpi2ps xmm1, mm2  
	
	movq	  mm1, [ebx + esi]
	
	movq	  mm2, mm1
	punpckhwd mm1, mm0  ; high
	cvtpi2ps xmm2, mm1 
	movlhps	 xmm2, xmm2
	punpcklwd mm2, mm0  ; low
	cvtpi2ps xmm2, mm2  
       
	mulps	xmm1, xmm2
	
	movaps	[edx + 2*esi], xmm1
	
	add	esi, 8
	cmp	esi, edi
	jl	.loop
	
	SWAPM  prev0, next0
	
	add esp, 4
	ret

	
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; compare_fields_mul_sse2
;
; edx = dest, esp = common data ptr
; will clobber all gp registers except ebp, esp, all mmx registers,
; and all sse registers except xmm7
;
	global _asm_compare_fields_mul_sse2
	align 16
_asm_compare_fields_mul_sse2: 
	push edx

	mov   eax, ebp	  
	mov   edx, [ebp + cur0]
	call  _compare_fields_row_isse
	
	mov   eax, ebp
	mov   edx, [ebp + next0]
	call  _compare_fields_row_isse
	
	mov edx, [esp]
	       
	XMMCON xmm0, 0x00000000
		
	mov  eax, [ebp + prev0]
	mov  ebx, [ebp + cur0]
	mov  ecx, [ebp + next0]
	mov  esi, 0
	mov  edi, [ebp + s_width]
	shl  edi, 1
	
	align16
.loop:
	movq	  xmm1, [eax + esi]
	movq	  xmm2, [ecx + esi]
	paddusw	  xmm1, xmm2
	punpcklwd xmm1, xmm0
	cvtdq2ps  xmm1, xmm1
	
	movq	  xmm2, [ebx + esi]
	punpcklwd xmm2, xmm0
	cvtdq2ps  xmm2, xmm2
       
	mulps	  xmm1, xmm2
	
	movaps	[edx + 2*esi], xmm1
	
	add	esi, 8
	cmp	esi, edi
	jl	.loop
	
	SWAPM  prev0, next0
	
	add esp, 4
	ret


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; compare_fields_row
;
; uses regparm(3) calling convention
;
; eax = common data ptr
; edx = dest
; will clobber ecx and all mmx registers
;
	global _compare_fields_row_isse
	align 16
_compare_fields_row_isse:
	sub esp, 24
	mov [esp +  8], ebx
	mov [esp + 12], esi
	mov [esp + 16], edi
	mov [esp + 20], ebp

	mov ebp, eax
	mov [esp + 0], edx
	mov [esp + 4], ebp
		    
	SWAPM  a, d
	SWAPM  b, e
	
	mov  esi, [ebp + line]
	and  esi, 1

	mov  eax, [ebp + c]
	call _asm_subsample_line
	
	mov  eax, [ebp + e]
	call _asm_subsample_line
 
	xor  esi, 1

	mov  eax, [ebp + d]
	call _asm_subsample_line

	mov	edi, [ebp + s_width]
	shl	edi, 1

	mov	esi, [esp +0] ; dest
	add	esi, edi
	push    esi

	mov	eax, [ebp+ a] ; eax = a [for inner loop]
	add	eax, edi
	mov	ebx, [ebp+ b] ; ebx = b [for inner loop]
	add	ebx, edi
	mov	ecx, [ebp+ c] ; ecx = c [for inner loop]
	add	ecx, edi
	mov	edx, [ebp+ d] ; edx = d [for inner loop]
	add	edx, edi
	mov	esi, [ebp+ e] ; esi = e [for inner loop]
	add	esi, edi

	neg	edi	  ; edi = x [for inner loop]

	call    [ebp + cf_row_loop]
	add     esp, 4

	mov	ebp, [esp + 4]
	inc	dword [ebp + line]
	
	mov ebx, [esp +	 8]
	mov esi, [esp + 12]
	mov edi, [esp + 16]
	mov ebp, [esp + 20]
	
	add	esp, 24
	ret
	

	global _asm_compare_fields_row_loop_isse
	align 16	
_asm_compare_fields_row_loop_isse:
	; loop prep
	MMCON mm5, 0x00010001, ebp
	MMCON mm6, 0x00030003, ebp

	mov ebp, [esp + 4]
	
	align16
.loop:
	movq	mm1, [ebx + edi] ; b  = B
	paddsw	mm1, [edx + edi] ; b += D
	pmullw	mm1, mm6	 ; b *= 3
	movq	mm2, [ecx + edi] ; c  = C
	psllw	mm2, 2		 ; c *= 4
	paddsw	mm2, [eax + edi] ; c += A
	paddsw	mm2, [esi + edi] ; c += E
	movq	mm3, mm2	 ; c2 = c
	psubusw mm2, mm1	 ; c = c - b (with unsigned sat)
	psubusw mm1, mm3	 ; b = b - c2 (with unsigned sat)
	por	mm1, mm2	 ; b = b ^ c
	paddusw mm1, mm5	 ; b += 1
	movq	[ebp + edi], mm1

	add	edi, 8
	jl	.loop

	ret


	global _asm_compare_fields_row_loop_isse2
	align 16
_asm_compare_fields_row_loop_isse2:
	; loop prep
	XMMCON xmm5, 0x00010001, ebp
	XMMCON xmm6, 0x00030003, ebp
	
	mov ebp, [esp + 4]

	align16
.loop:
	movdqa	xmm1, [ebx + edi] ; b  = B
	paddsw	xmm1, [edx + edi] ; b += D
	pmullw	xmm1, xmm6	  ; b *= 3
	movdqa	xmm2, [ecx + edi] ; c  = C
	psllw	xmm2, 2		  ; c *= 4
	paddsw	xmm2, [eax + edi] ; c += A
	paddsw	xmm2, [esi + edi] ; c += E
	movdqa	xmm3, xmm2	  ; c2 = c
	psubusw xmm2, xmm1	  ; c = c - b (with unsigned sat)
	psubusw xmm1, xmm3	  ; b = b - c2 (with unsigned sat)
	por	xmm1, xmm2	  ; b = b ^ c
	paddusw xmm1, xmm5	  ; b += 1
	movdqa	[ebp + edi], xmm1

	add	edi, 16
	jl	.loop

	ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; memcopy
;
; eax = src
; ebx = stop (for src)
; edx = dest
;
; stop - start % 4 == 0
;
; clobbers ecx
;
	align 16
_asm_memcopy:
	 mov ecx, [eax]
	 mov [edx], ecx
	 add eax, 4
	 add edx, 4
	 cmp eax,ebx
	 jl _asm_memcopy
	 ret


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; _memclear
;
; eax = start, edx = stop
;
; stop - start % 8 == 0
;
	global _memclear
	align 16
_memclear:
	mov  dword [eax],     0
	mov  dword [eax + 4], 0
	add eax, 8
	cmp eax, edx
	jl  _memclear
	ret
